/******************************************************************************
 * $Id: guild_parser.c,v 1.2 2001/04/21 02:56:06 mud Exp $
 *   
 * File: guild_parser.c
 *
 * Created: Sat May 13 2000 by Yohay Etsion (yohay_e@netvision.net.il)
 *
 * This is part of Dragons Fang, which is based on CircleMUD, etc, etc, yada, 
 * yada...
 *
 * A nice description of the files general purpose would be nice here... :)
 * 
 *****************************************************************************/

#include "gnome-xml/parser.h"
#include "gnome-xml/tree.h"
#include "gnome-xml/entities.h"
#include "gnome-xml/parserInternals.h"
#include "stdio.h"

#include "conf.h"
#include "sysdep.h"

#include "guild.h"
#include "guild_parser.h"
#include "structs.h"
#include "utils.h"
#include "comm.h"
#include "interpreter.h"
#include "handler.h"
#include "db.h"
#include "spells.h"
#include "screen.h"
#include "constants.h"
#include "boards.h"

/****** Globals ***************/

struct guild_info *guilds_data;

/****** XML Utilities *********/

// Mainly contributed by Stefan Lindstrom :)

char *xmlToString(xmlNodePtr p)
{
  xmlNodePtr temp;
  char *target;
  int size = 0;

  for (temp = p; NULL != temp; temp = temp->next) {
    size += strlen(temp->content);
  }
  size += 1;

  target = calloc(1,size);
  *target = '\0';

  for (temp = p; NULL != temp; temp = temp->next) {
    strcat(target,temp->content);
  }
  if (strlen(target) >= MAX_STRING_LENGTH) {
    free(target);
    return (strdup("STRING TO LONG!\r\n"));
  }
  // The gnome XML parser (according to a od -cx at least :))
  // preserves the \r\n sequence of our descriptions, hence, we
  // can just return our string here...

  return target;
}

/*
 * Kinda unescessary wrapping, but since I use this func in a lot of places...
 */
char *internalXmlString(char *old, xmlDocPtr doc)
{
  return (xmlEncodeEntitiesReentrant(doc,old));
}

/****** XML savings *******/

static xmlNodePtr makeRankInfo(struct rank_info *r, xmlDocPtr doc)
{
  char *p, buf[1024];
  xmlNodePtr rank = xmlNewNode(NULL,"rank");
  rank->doc = doc;

  xmlNewProp(rank,"name",(p = internalXmlString(r->name,doc)));
  free(p);
 
  sprintf(buf, "%ld", r->num);
  xmlNewProp(rank, "num", buf);

  return rank;
}

static xmlNodePtr makeSponsorerInfo(struct sponsorer_info *s, xmlDocPtr doc)
{
  char *p, buf[1024];
  xmlNodePtr sponsorer = xmlNewNode(NULL, "sponsorer");
  sponsorer->doc = doc;

  sprintf(buf, "%ld", s->idnum);
  xmlNewProp(sponsorer, "idnum", buf);

  xmlNewProp(sponsorer, "name", (p = internalXmlString(s->name,doc)));
  free(p);

  return sponsorer;
}

static xmlNodePtr makeGequipInfo(struct gequip_info *g, xmlDocPtr doc)
{
  char buf[1024];
  xmlNodePtr gequip = xmlNewNode(NULL,"gequip");
  gequip->doc = doc;

  sprintf(buf, "%d", g->vnum);
  xmlNewProp(gequip,"vnum", buf);

  return gequip;
}

static xmlNodePtr makeGuildieInfo(struct guildie_info *g, xmlDocPtr doc)
{
  char *p, buf[1024];
  xmlNodePtr guildie = xmlNewNode(NULL,"guildie");
  struct sponsorer_info *s;
  struct gequip_info *gequip;
  guildie->doc = doc;

  sprintf(buf, "%ld", g->idnum);
  xmlNewProp(guildie,"idnum", buf);

  xmlNewProp(guildie,"name",(p = internalXmlString(g->name,doc)));
  free(p);
 
  sprintf(buf, "%ld", g->rank_num);
  xmlNewProp(guildie,"rank_num", buf);

  xmlNewProp(guildie,"subrank",(p = internalXmlString(g->subrank,doc)));
  free(p);

  sprintf(buf, "%ld", g->perm);
  xmlNewProp(guildie,"perm", buf);

  sprintf(buf, "%ld", g->status);
  xmlNewProp(guildie,"status", buf);

  sprintf(buf, "%ld", g->deposited);
  xmlNewProp(guildie,"deposited", buf);

  sprintf(buf, "%ld", g->withdrew);
  xmlNewProp(guildie,"withdrew", buf);

  s = g->sponsorers;

  while (s) {
   xmlAddChild(guildie, makeSponsorerInfo(s, doc));
   s = s->next;
  }

  gequip = g->gequipsent;
  while (gequip) {
   xmlAddChild(guildie, makeGequipInfo(gequip, doc));
   gequip = gequip->next;
  }

  return guildie;
}

static xmlNodePtr makeGskillInfo(struct gskill_info *g, xmlDocPtr doc)
{
  char buf[1024];
  xmlNodePtr gskill = xmlNewNode(NULL,"gskill");
  gskill->doc = doc;

  sprintf(buf, "%d", g->skill);
  xmlNewProp(gskill,"skill", buf);

  sprintf(buf, "%d", g->maximum_set);
  xmlNewProp(gskill,"maximum_set", buf);
  
  return gskill;
}

static xmlNodePtr makeGzoneInfo(struct gzone_info *g, xmlDocPtr doc)
{
  char buf[1024];
  xmlNodePtr gzone = xmlNewNode(NULL,"gzone");
  gzone->doc = doc;

  sprintf(buf, "%d", g->zone);
  xmlNewProp(gzone,"zone", buf);
  
  return gzone;
}

static xmlNodePtr makeGhelpInfo(struct ghelp_info *g, xmlDocPtr doc)
{
  char *p;
  xmlNodePtr ghelp = xmlNewNode(NULL, "ghelp");
  ghelp->doc = doc;

  xmlNewProp(ghelp, "keyword", g->keyword);

  xmlNewChild(ghelp,NULL,"entry",(p = internalXmlString(g->entry,doc)));
  free(p);

  return ghelp;
}

static xmlNodePtr makeGuildBasic(struct guild_info *g,xmlDocPtr doc)
{
  char buf[1024],*p;
  xmlNodePtr basic = xmlNewNode(NULL,"guildbasic");
  basic->doc = doc;

  xmlNewProp(basic,"name",g->name);
  xmlNewProp(basic,"guildfile",g->guild_filename);
  xmlNewProp(basic,"gossip_name",g->gossip_name);

  sprintf(buf, "%d", g->type);
  xmlNewProp(basic,"type", buf);

  sprintf(buf, "%d", g->id);
  xmlNewProp(basic,"id", buf);

  xmlNewProp(basic, "gltitle", (p = internalXmlString(g->gl_title,doc)));
  free(p);

  xmlNewProp(basic, "guildietitles", (p = internalXmlString(g->guildie_titles,doc)));
  free(p);

  sprintf(buf, "%ld", g->gflags);
  xmlNewProp(basic,"gflags", buf);
  
  sprintf(buf, "%d", g->guildwalk_room);
  xmlNewProp(basic,"guildwalk_room", buf);

  xmlNewChild(basic,NULL,"desc",(p = internalXmlString(g->description,doc)));
  free(p);

  xmlNewChild(basic,NULL,"reqs",(p = internalXmlString(g->requirements,doc)));
  free(p);

  xmlNewChild(basic,NULL,"gossip",(p = internalXmlString(g->gossip,doc)));
  free(p);

  xmlNewChild(basic,NULL,"gchan",(p = internalXmlString(g->gchan_name,doc)));
  free(p);
  
  xmlNewProp(basic->childs->next->next->next, "gchan_color", (p = internalXmlString(g->gchan_color,doc)));
  free(p);

  sprintf(buf, "%d", g->gchan_type);
  xmlNewProp(basic->childs->next->next->next, "gchan_type", buf);

  sprintf(buf, "%d", g->gold);
  xmlNewProp(basic, "gold", buf);

  return basic;
}

void GuildToXml(struct guild_info *g, FILE *f)
{
  xmlNodePtr guild;
  xmlDocPtr doc;
  struct gskill_info *gskills;
  struct guildie_info *guildies;
  struct gzone_info *gzones;
  struct rank_info *ranks;
  struct gequip_info *gequip;
  struct ghelp_info *ghelp;

  doc = xmlNewDoc("1.0");
  guild = xmlNewDocNode(doc, NULL, "guildfile", NULL);
  doc->root = guild;

  xmlAddChild(guild, makeGuildBasic(g, doc));

  gskills = g->gskills;
  guildies = g->guildies;
  gzones = g->gzones;
  ranks = g->ranks;
  gequip = g->gequip;
  ghelp = g->ghelp;

  while (ranks) {
	  xmlAddChild(guild, makeRankInfo(ranks, doc));
	  ranks = ranks->next;
  }

  while (gskills) {
	  xmlAddChild(guild, makeGskillInfo(gskills, doc));
	  gskills = gskills->next;
  }

  while (gzones) {
	  xmlAddChild(guild, makeGzoneInfo(gzones, doc));
	  gzones = gzones->next;
  }

  while (guildies) {
	  xmlAddChild(guild, makeGuildieInfo(guildies,doc));
	  guildies = guildies->next;
  }

  while (gequip) {
	  xmlAddChild(guild, makeGequipInfo(gequip, doc));
	  gequip = gequip->next;
  }

  while (ghelp) {
   	  xmlAddChild(guild, makeGhelpInfo(ghelp, doc));
 	  ghelp = ghelp->next;
  }

  xmlDocDump(f, doc);
  xmlFreeDoc(doc); 
}


/******** XML loadings *********/

static void fetchRankInfo(xmlNodePtr rank, struct rank_info *r)
{
  r->num = atoi(xmlGetProp(rank, "num"));
  r->name = strdup(xmlGetProp(rank, "name"));
}

static void fetchSponsorerInfo(xmlNodePtr sponsorer, struct sponsorer_info *s)
{
  s->idnum = atoi(xmlGetProp(sponsorer, "idnum"));
  s->name = strdup(xmlGetProp(sponsorer, "name"));
}

static void fetchGequipInfo(xmlNodePtr gequip, struct gequip_info *g)
{
  g->vnum = atoi(xmlGetProp(gequip, "vnum"));
}

static void fetchGuildieInfo(xmlNodePtr guildie, struct guildie_info *g)
{
  struct sponsorer_info *s;
  struct gequip_info *gequip;
  xmlNodePtr temp;

  g->idnum = atoi(xmlGetProp(guildie, "idnum"));
  g->name = strdup(xmlGetProp(guildie, "name"));
  g->rank_num = atoi(xmlGetProp(guildie, "rank_num"));
  g->subrank = strdup(xmlGetProp(guildie, "subrank"));
  g->perm = atoi(xmlGetProp(guildie, "perm"));
  g->status = atoi(xmlGetProp(guildie, "status"));
  g->deposited = atoi(xmlGetProp(guildie, "deposited"));
  g->withdrew = atoi(xmlGetProp(guildie, "withdrew"));

  for (temp = guildie->childs; NULL != temp; temp = temp->next) {
   if (!strcasecmp(temp->name, "SPONSORER")) {
    s = calloc(1, sizeof(struct sponsorer_info));
    s->next = g->sponsorers;
    g->sponsorers = s;
    fetchSponsorerInfo(temp, s);
   }
   if (!strcasecmp(temp->name, "GEQUIP")) {
    gequip = calloc(1, sizeof(struct gequip_info));
    gequip->next = g->gequipsent;
    g->gequipsent = gequip;
    fetchGequipInfo(temp, gequip);
   } 
  }    
}

static void fetchGskillInfo(xmlNodePtr gskill, struct gskill_info *g)
{
  g->skill = atoi(xmlGetProp(gskill, "skill"));
  g->maximum_set = atoi(xmlGetProp(gskill, "maximum_set"));
}

static void fetchGzoneInfo(xmlNodePtr gzone, struct gzone_info *g)
{
  g->zone = atoi(xmlGetProp(gzone, "zone"));
}

static void fetchGhelpInfo(xmlNodePtr ghelp, struct ghelp_info *g)
{
  g->keyword = strdup(xmlGetProp(ghelp, "keyword"));
  g->entry = xmlToString(ghelp->childs->childs);
}

static void fetchGuildBasics(xmlNodePtr basic, struct guild_info *g)
{
  xmlNodePtr temp;

  g->name = strdup(xmlGetProp(basic, "name"));
  g->guild_filename = strdup(xmlGetProp(basic, "guildfile"));
  g->gossip_name = strdup(xmlGetProp(basic, "gossip_name"));
  g->gl_title = strdup(xmlGetProp(basic, "gltitle"));
  g->guildie_titles = strdup(xmlGetProp(basic, "guildietitles"));
  g->type = atoi(xmlGetProp(basic, "type"));
  g->id = atoi(xmlGetProp(basic, "id"));
  g->gflags = atoi(xmlGetProp(basic, "gflags"));
  g->guildwalk_room = atoi(xmlGetProp(basic, "guildwalk_room"));
  g->gold = atoi(xmlGetProp(basic, "gold"));

  for (temp = basic->childs; temp != NULL; temp = temp->next) {
    if (!strcasecmp(temp->name, "desc"))
     g->description = xmlToString(temp->childs);

    else if (!strcasecmp(temp->name, "reqs"))
     g->requirements = xmlToString(temp->childs);

    else if (!strcasecmp(temp->name, "gossip"))
     g->gossip = xmlToString(temp->childs);

    else if (!strcasecmp(temp->name, "gchan")) {
     g->gchan_name = xmlToString(temp->childs);
     g->gchan_color = strdup(xmlGetProp(temp, "gchan_color"));
     g->gchan_type = atoi(xmlGetProp(temp, "gchan_type"));
    }

    else 
     continue;
  }
}

void XmlToGuild(char *file)
{
   xmlNodePtr root, temp;
   xmlDocPtr doc;
   struct guild_info *g;
   struct gequip_info *gequip;
   struct gskill_info *gskill;
   struct guildie_info *guildie;
   struct gzone_info *gzone;
   struct rank_info *rank, *rank2;
   struct ghelp_info *ghelp;
 
   if (!(doc = xmlParseFile(file))) {
    log("XmlToGuild:[guild_parser.c]: Couldn't parse XML-file %s",file);
    return;
   }
   if (!(root = doc->root)) {
    log("XmlToGuild:[guild_parser.c]: No XML-root in document %s",file);
    return;
   }
   if (strncasecmp("GUILDFILE",root->name,strlen(root->name))) {
    log("XmlToGuild:[guild_parser.c]: XML-root (%s in file %s) is not a guild-object!",root->name,file);
    return;
   }

   g = calloc(1, sizeof(struct guild_info));
   g->next = guilds_data;
   guilds_data = g;

   for (temp = root->childs; NULL != temp; temp = temp->next) {

    if (!strcasecmp(temp->name, "GUILDBASIC")) 
	fetchGuildBasics(temp, g);

    else if (!strcasecmp(temp->name, "RANK")) {
          rank = calloc(1, sizeof(struct rank_info));
	  if (!g->ranks) {
	   g->ranks = rank;
	  }
	  else {
	   rank2 = g->ranks;
	   while (rank2->next != NULL)
	    rank2 = rank2->next;
	   rank2->next = rank;
 	  }
 	  fetchRankInfo(temp, rank);
    }	

    else if (!strcasecmp(temp->name, "GUILDIE")) { 
          guildie = calloc(1, sizeof(struct guildie_info));
	  guildie->next = g->guildies;
	  g->guildies = guildie;
	  fetchGuildieInfo(temp, guildie);
          rank = g->ranks;
          while (rank) 
            if (guildie->rank_num == rank->num) {
             guildie->rank = rank;
             rank = 0;
            }
            else rank = rank->next;
    }

    else if (!strcasecmp(temp->name, "GSKILL"))  {
	  gskill = calloc(1, sizeof(struct gskill_info));
	  gskill->next = g->gskills;
	  g->gskills = gskill;
	  fetchGskillInfo(temp, gskill);
    }

    else if (!strcasecmp(temp->name, "GZONE"))   {
	  gzone = calloc(1, sizeof(struct gzone_info));
	  gzone->next = g->gzones;
 	  g->gzones = gzone;
	  fetchGzoneInfo(temp, gzone);
    }

    else if (!strcasecmp(temp->name, "GEQUIP"))  {
	  gequip = calloc(1, sizeof(struct gequip_info));
	  gequip->next = g->gequip;
	  g->gequip = gequip;
	  fetchGequipInfo(temp, gequip);
    }
 
    else if (!strcasecmp(temp->name, "GHELP"))  {
 	  ghelp = calloc(1, sizeof(struct ghelp_info));
	  ghelp->next = g->ghelp;
	  g->ghelp = ghelp;
	  fetchGhelpInfo(temp, ghelp);
    }
   }
   xmlFreeDoc(doc);
}
